/****************************************************************************/
// camcam.c
//
//  - routines that connect the GUI with the USB functionality
//  - camera disconnected/reconnected notifications
//  - secondary thread that handles most camera operations
//
/****************************************************************************/

/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Cameraderie.
 *
 * The Initial Developer of the Original Code is Richard L. Walsh.
 * 
 * Portions created by the Initial Developer are Copyright (C) 2006
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * ***** END LICENSE BLOCK ***** */

/****************************************************************************/

#include "camapp.h"
#include "camusb.h"
#define INCL_DOS
#define INCL_WIN
#include <os2.h>

/****************************************************************************/

enum    eNotify     { eSEMADD, eSEMREMOVE, eSEMTERM, eCNTSEMS, eSEMERR};

typedef struct _CAMNotify {
    SEMRECORD   sr[eCNTSEMS];
    CAMDevice   device;
    HMUX        hmux;
    HWND        hwnd;
    uint32_t    NotifyID;
} CAMNotify;

/****************************************************************************/

/*
ULONG           SetupCamera( HMQ hmq);
void            ShutDownCamera( void);
ULONG           InitNotification( HWND hwnd);
void        	HandleNotification( HWND hwnd, MPARAM mp1, MPARAM mp2);
ULONG           DispatchUsbMsg( ULONG ulMsg, ULONG ulParam, PVOID * ppRtn);
*/

ULONG           ReconnectCamera( CAMDevice device);
void            NotifyThreadProc( void * arg);
void APIENTRY   TermNotifyOnExit( ULONG ul);
ULONG           InitUsbThread( void);
MRESULT _System ReplyWndProc( HWND hwnd, ULONG msg, MPARAM mp1, MPARAM mp2);
void            UsbThreadProc( void * arg);
MRESULT _System UsbWndProc( HWND hwnd, ULONG msg, MPARAM mp1, MPARAM mp2);
void            ResetPriority( void);

uint32_t        SaveObject( CAMHandle hCam, CAMSaveObj * cso, MPARAM mp2);

/****************************************************************************/

static BOOL             fInUse = FALSE;
static BOOL             fStop = TRUE;
static CAMDeviceInfoPtr pDevInfo = 0;
static CAMHandle        hCam = 0;
static HEV              hevTerm = 0;
static HWND             hReply = 0;
static HWND             hUsb = 0;
static ULONG            ulNotifyID = 0;

/****************************************************************************/

ULONG       SetupCamera( HMQ hmq)

{
    ULONG   	     rc;
    ULONG            ctr;
    CAMCameraListPtr pList = 0;
    CAMDevice        device;

do
{
    SetUsbDebug( 2);

    rc = InitUsb();
    if (rc)
        break;

    rc = InitUsbThread();
    if (rc)
        break;

    rc = GetCameraList( &pList, (uint32_t)CAM_FORCE_FLAG);

    while (rc == CAMERR_NOCAMERAS) {
        if (ShowInitErrorDlg( 0, IDD_NOCAM) != IDC_RETRY)
            break;

        rc = GetCameraList( &pList, (uint32_t)CAM_FORCE_FLAG);
    }
    if (rc)
        break;

    // if the force flag is off, there's only 1 entry in the list;
    // if it's on, try all found devices
    for (ctr = 0; ctr < pList->cntCameras; ctr++) {
        device = pList->Cameras[ctr].Device;
        rc = InitCamera( device, &hCam);
        if (!rc)
            break;
    }
    if (rc)
        break;

    rc = GetDeviceInfo( hCam, &pDevInfo);
    if (rc)
        break;

} while (0);

    // if there was an error, abort
    if (rc && rc != CAMERR_NOCAMERAS)
        ShowInitErrorDlg( MSG_InitCamera, IDD_ERROR);

    free( pList);

    return rc;
}

/****************************************************************************/

void        ShutDownCamera( void)

{
    free( pDevInfo);
    pDevInfo = 0;
    if (hevTerm) {
        DosPostEventSem( hevTerm);
        DosSleep( 30);
    }
    TermCamera( &hCam);
    TermUsb();

    return;
}

/****************************************************************************/

ULONG       ReconnectCamera( CAMDevice device)

{
    ULONG   	 rc;

do {
    if (!device) {
        rc = CAMERR_NOCAMERAS;
        break;
    }

    rc = InitCamera( device, &hCam);
    if (rc)
        break;

    rc = GetDeviceInfo( hCam, &pDevInfo);
    if (rc)
        break;

} while (0);

    return rc;
}

/****************************************************************************/
/****************************************************************************/

ULONG       InitNotification( HWND hwnd)

{
    APIRET      rc = 0;
    CAMNotify * pCN = 0;

do {
    pCN = malloc( sizeof(CAMNotify));
    if (!pCN) {
        rc = CAMERR_MALLOC;
        break;
    }
    memset( pCN, 0, sizeof(CAMNotify));

    pCN->hwnd = hwnd;
    pCN->device = GetDevice( hCam);
    if (!pCN->device) {
        rc = CAMERR_MISCELLANEOUS;
        break;
    }

    pCN->sr[eSEMADD].ulUser = eSEMADD;
    rc = DosCreateEventSem( 0, (PHEV)&(pCN->sr[eSEMADD].hsemCur), DC_SEM_SHARED, 0);
    if (rc)
        break;

    pCN->sr[eSEMREMOVE].ulUser = eSEMREMOVE;
    rc = DosCreateEventSem( 0, (PHEV)&(pCN->sr[eSEMREMOVE].hsemCur), DC_SEM_SHARED, 0);
    if (rc)
        break;

    pCN->sr[eSEMTERM].ulUser = eSEMTERM;
    rc = DosCreateEventSem( 0, (PHEV)&(pCN->sr[eSEMTERM].hsemCur), 0, 0);
    if (rc)
        break;

    hevTerm = (HEV)pCN->sr[eSEMTERM].hsemCur;

    rc = DosCreateMuxWaitSem( 0, &pCN->hmux, eCNTSEMS, pCN->sr, DCMW_WAIT_ANY);
    if (rc)
        break;

    rc = StartNotification( pCN->device, (HEV)pCN->sr[eSEMADD].hsemCur,
                            (HEV)pCN->sr[eSEMREMOVE].hsemCur, &pCN->NotifyID);
    if (rc)
        break;

    ulNotifyID = pCN->NotifyID;
    rc = DosExitList( ((0x7f << 8) | EXLST_ADD), TermNotifyOnExit);
    if (rc) {
        camcli_error( "InitNotification:  DosExitList ADD failed - rc= %d", (int)rc);
        rc = 0;
    }

    if (_beginthread( NotifyThreadProc, 0, 0x4000, pCN) == -1) {
        rc = CAMERR_MISCELLANEOUS;
        break;
    }

} while (0);

    if (rc) {
        camcli_error( "InitNotification - rc= %d\n", rc);

        if (pCN) {
            if (pCN->NotifyID)
                EndNotification( pCN->NotifyID);
            if (pCN->hmux)
                DosCloseMuxWaitSem( pCN->hmux);
            if (pCN->sr[eSEMTERM].hsemCur)
                DosCloseEventSem( (HEV)pCN->sr[eSEMTERM].hsemCur);
            if (pCN->sr[eSEMREMOVE].hsemCur)
                DosCloseEventSem( (HEV)pCN->sr[eSEMREMOVE].hsemCur);
            if (pCN->sr[eSEMADD].hsemCur)
                DosCloseEventSem( (HEV)pCN->sr[eSEMADD].hsemCur);
            free( pCN);
        }
        hevTerm = 0;
    }

    return (rc);
}

/****************************************************************************/

void        NotifyThreadProc( void * arg)

{
    APIRET      rc;
    CAMNotify * pCN = (CAMNotify*)arg;
    ULONG       ulSem;
    ULONG       ulCnt;

    if (!arg)
        return;

    DosResetEventSem( (HEV)pCN->sr[eSEMADD].hsemCur, &ulCnt);

    while (1) {
        rc = DosWaitMuxWaitSem( pCN->hmux, (ULONG)-1, &ulSem);
        if (rc)
            ulSem = eSEMERR;

        if (ulSem < eCNTSEMS) {
            ulCnt = 0;
            rc = DosResetEventSem( (HEV)pCN->sr[ulSem].hsemCur, &ulCnt);
            if (ulCnt != 1)
                camcli_error( "Multiple event sem postings - sem= %d  cnt= %d",
                              (int)ulSem, (int)ulCnt);
            if (rc)
                ulSem = eSEMERR;
        }
        else
            ulSem = eSEMERR;

        if (ulSem != eSEMTERM)
            WinPostMsg( pCN->hwnd, CAMMSG_NOTIFY, (MP)ulSem, (MP)pCN->device);

        if (rc || (ulSem != eSEMADD && ulSem != eSEMREMOVE))
            break;
    }

    if (rc)
        camcli_error( "exiting NotifyThreadProc - rc= %d", (int)rc);

    rc = EndNotification( pCN->NotifyID);
    if (rc)
        camcli_error( "EndNotification - rc= %x", (int)rc);
    else {
        ulNotifyID = 0;
        rc = DosExitList( EXLST_REMOVE, TermNotifyOnExit);
        if (rc)
            camcli_error( "DosExitList REMOVE failed - rc= %d", (int)rc);
    }

    DosCloseMuxWaitSem( pCN->hmux);
    hevTerm = 0;
    DosCloseEventSem( (HEV)pCN->sr[eSEMTERM].hsemCur);
    DosCloseEventSem( (HEV)pCN->sr[eSEMREMOVE].hsemCur);
    DosCloseEventSem( (HEV)pCN->sr[eSEMADD].hsemCur);
    free( pCN);

    return;
}

/****************************************************************************/

// this should clean up in case we crash

void APIENTRY   TermNotifyOnExit( ULONG ul)

{
    if (ulNotifyID) {
        APIRET rc = EndNotification( ulNotifyID);
        printf( "TermNotifyOnExit:  TermNotification - rc= %x\n", (int)rc);
        ulNotifyID = 0;
    }

    DosExitList( EXLST_EXIT, 0);
    return;
}

/****************************************************************************/

void        HandleNotification( HWND hwnd, MPARAM mp1, MPARAM mp2)

{
static  HWND    hDisconnect = 0;
        ULONG   rc;

    if (mp1 == (MP)eSEMREMOVE) {
        if (!hDisconnect)
            hDisconnect = OpenWaitOrExitDlg( hwnd, IDD_DISCONNECT);

        free( pDevInfo);
        pDevInfo = 0;
        AbandonCamera( &hCam);

        return;
    }

    if (mp1 == (MP)eSEMADD) {
        rc = ReconnectCamera( (CAMDevice)mp2);
        CloseWaitOrExitDlg( hDisconnect);
        hDisconnect = 0;

        if (!rc)
            WinPostMsg( hwnd, WM_COMMAND, (MP)IDM_SYNCDATA, 0);
        else {
            ShowInitErrorDlg( MSG_ReopenCamera, IDD_ERROR);
            WinPostMsg( hwnd, WM_QUIT, 0, 0);
        }

        return;
    }

    if (mp2) {
        if (InitNotification( hwnd))
            WinAlarm( HWND_DESKTOP, WA_ERROR);
    }

    return;
}

/****************************************************************************/
/****************************************************************************/

// this creates a thread that will handle time-consuming camera operations
// such as fetching pictures;  it also creates a window that the camera
// thread will post its replies to - that window should never actually
// receive those replies because they should be captured in DispatchUsbMsg()'s
// msg loop

ULONG       InitUsbThread( void)

{
    APIRET  rc = CAMERR_DISPATCHINIT;
    ULONG   ctr;

do {

    // create a static-class window 'cause it's easy
    hReply = WinCreateWindow( HWND_OBJECT, WC_STATIC, "",
                              SS_TEXT, 0, 0, 0, 0, 0, HWND_BOTTOM,
                              1234, 0, 0);
    if (hReply == 0)
        break;

    // subclass it - the subclass proc will completely
    // replace the original wndproc
    if (WinSubclassWindow( hReply, ReplyWndProc) == 0)
        break;

    // start the camera thread
    if (_beginthread( UsbThreadProc, 0, 0xF000, 0) == -1)
        break;

    // wait a while for the camera thread to create the
    // window to which DispatchUsbMsg() will post its requests
    for (ctr = 0; ctr < 15; ctr++) {
        DosSleep( 64);
        if (hUsb)
            break;
    }

    // if the window was created, we're OK
    if (hUsb)
        rc = 0;

} while (0);

    if (rc) {
        camcli_error( "InitUsbThread - hUsb= %x  hReply= %x",
                      (int)hUsb, (int)hReply);
        if (hReply)
            WinDestroyWindow( hReply);
        hReply = 0;
    }

    return rc;
}

/****************************************************************************/

// This dispatches calls "pseudo-synchronously".  It posts a msg to the
// Usb thread's window, then enters a msg loop.  Because the msgs it
// dispatches while waiting could cause it to be called recursively,
// it has to be made non-reentrant via the fInUse flag.  Without it,
// the last invocation would take control, discard replies to earlier
// requests, and cause earlier invocations to pile up on the stack.
//
// The function exits when the reply is received, a timeout occurs,
// or WM_QUIT is received;  if it receives a reply to a previous msg
// which timed-out, it frees that msg's payload (if any) then continues
// waiting.

ULONG       DispatchUsbMsg( ULONG ulMsg, ULONG ulParam, void** ppRtn)

{
    ULONG       rc = 0;
    ULONG       ulTimer = 0;
    ULONG       ulMsgId = 0;
    QMSG        qmsg;

do
{
    if (CAM_DEMO_FLAG)
        ERRNBR( CAMERR_DEMOMODE)

    if (ppRtn)
        *ppRtn = 0;

    if (fInUse)
        ERRNBR( CAMERR_DISPATCHINUSE)

    if (hUsb == 0)
        ERRNBR( CAMERR_DISPATCHERR)

    // use the current time as a msg ID
    DosQuerySysInfo( QSV_MS_COUNT, QSV_MS_COUNT, &ulMsgId, sizeof(ulMsgId));

    // start the timeout timer
    ulTimer = WinStartTimer( 0, hReply, (ULONG)TID_USERMAX-27, 10000);
    if (ulTimer == 0)
        ERRNBR( CAMERR_DISPATCHERR)

    fStop = FALSE;

    if (WinPostMsg( hUsb, ulMsg, (MP)ulParam, (MP)ulMsgId) == FALSE)
        ERRNBR( CAMERR_DISPATCHERR)

    // set the fInUse flag before we enter the msg loop
    fInUse = TRUE;

    // this is effectively a "do-forever" loop
    while (rc == 0)
    {
        // WM_QUIT - repost msg then exit without resetting fInUse
        if (WinGetMsg( 0, &qmsg, 0, 0, 0) == FALSE)
        {
            WinPostMsg( qmsg.hwnd, qmsg.msg, qmsg.mp1, qmsg.mp2);
            ERRNBR( CAMERR_DISPATCHQUIT)
        }

        // handle all msgs except the server's reply;
        // if it's our WM_TIMER msg, reset fInUse then exit;
        // otherwise dispatch the msg
        if (qmsg.hwnd != hReply || qmsg.msg != CAMMSG_REPLY)
        {
            if (qmsg.msg == CAMMSG_KEEPALIVE) {
                if ((ULONG)qmsg.mp2 == ulMsgId)
                    ulTimer = WinStartTimer( 0, hReply,
                                             (ULONG)TID_USERMAX-27, 10000);
                continue;
            }

            if (qmsg.msg == WM_TIMER &&
                SHORT1FROMMP( qmsg.mp1) == (USHORT)ulTimer) {
                    fInUse = FALSE;
                    ERRNBR( CAMERR_DISPATCHTIMEOUT)
            }

            if (qmsg.msg == WM_COMMAND) {
                if (SHORT1FROMMP( qmsg.mp1) == IDC_STOP)
                    qmsg.mp1 = (MP)IDC_HALT;
                if (SHORT1FROMMP( qmsg.mp1) == IDC_HALT)
                    fStop = TRUE;
                qmsg.mp2 = (MP)-1;
            }

            WinDispatchMsg( 0, &qmsg);
            continue;
        }

        // this is the msg we're expecting - store the return
        // value or error code, reset fInUse, then exit
        if ((ULONG)qmsg.mp2 == ulMsgId)
        {
            if ((ULONG)qmsg.mp1 > 0x10000) {
                if (ppRtn)
                    *ppRtn = qmsg.mp1;
            }
            else
                rc = (ULONG)qmsg.mp1;
            fInUse = FALSE;
            break;
        }
        else {
            // if the msgID is less than the current value, this is
            // the reply to a request that timed out, so free its
            // payload & keep waiting;  if it's greater, something
            // wierd is going on, so reset fInUse then exit
            if ((ULONG)qmsg.mp1 > 0x10000)
                free( qmsg.mp1);
            if ((ULONG)qmsg.mp2 > ulMsgId) {
                rc = CAMERR_DISPATCHUNEXPECTED;
                fInUse = FALSE;
                break;
            }
        }
    }

} while (0);

    // if the timer is on & the function is no longer in use, turn it off
    if (!fInUse && ulTimer)
        WinStopTimer( 0, hReply, ulTimer);

    fStop = TRUE;

    if (rc)
        camcli_error( "DispatchUsbMsg - rc= %x", (int)rc);

    return (rc);
}

/****************************************************************************/

// This proc completely overrides WC_STATIC's native msg handling.
// The window it supports is intended primarily as a marker.  Although
// the Usb thread posts reply msgs to it, they should be intercepted in
// DispatchUsbMsg()'s msg loop and never reach this wndproc.  If they do,
// then there was an error elsewhere and this wndproc does cleanup.

MRESULT _System ReplyWndProc( HWND hwnd, ULONG msg, MPARAM mp1, MPARAM mp2)

{
    if (msg == CAMMSG_REPLY)
    {
        camcli_error( "ReplyWndProc:  CAMMSG_REPLY received - mp1= %x",
                      (int)mp1);
        if ((ULONG)mp1 > 0x10000)
            free( mp1);
        return (0);
    }

    if (msg == WM_TIMER && SHORT1FROMMP( mp1) == TID_USERMAX-27)
    {
        camcli_error( "ReplyWndProc:  WM_TIMER received");
        WinStopTimer( 0, hwnd, SHORT1FROMMP( mp1));
        return (0);
    }

    if (msg == WM_DESTROY)
        hReply = 0;

    return (WinDefWindowProc( hwnd, msg, mp1, mp2));
}

/****************************************************************************/

void        UsbThreadProc( void * arg)

{
    HAB     hab = 0;
    HMQ     hmq = 0;
    QMSG    qmsg;

do
{
    // ensure this thread preempts the UI thread
    ResetPriority();

    hab = WinInitialize( 0);
    if (!hab)
        break;

    hmq = WinCreateMsgQueue( hab, 0);
    if (!hmq)
        break;

    // create then subclass the window requests will be posted to
    hUsb = WinCreateWindow( HWND_OBJECT, WC_STATIC, "",
                            SS_TEXT, 0, 0, 0, 0, 0, HWND_BOTTOM,
                            5678, 0, 0);
    if (hUsb == 0)
        break;

    if (WinSubclassWindow( hUsb, UsbWndProc) == 0)
        break;

    while (WinGetMsg( hab, &qmsg, NULLHANDLE, 0, 0))
        WinDispatchMsg( hab, &qmsg);

} while (0);

    if (hUsb) {
        WinDestroyWindow( hUsb);
        hUsb = 0;
    }
    if (hmq)
        WinDestroyMsgQueue( hmq);
    if (hab)
        WinTerminate( hab);

    return;
}

/****************************************************************************/

MRESULT _System UsbWndProc( HWND hwnd, ULONG msg, MPARAM mp1, MPARAM mp2)

{
    APIRET  rc;
    MPARAM  mpRtn;
    void *  pData = 0;

    switch (msg) {

        case CAMMSG_GETOBJECTHANDLES:
            rc = GetObjectHandles( hCam, 0xffffffff, 0, 0,
                                   (CAMObjectHandlePtr*)&pData);
            break;

        case CAMMSG_GETOBJECTINFO:
            rc = GetObjectInfo( hCam, (ULONG)mp1,
                                (CAMObjectInfoPtr*)&pData);
            break;

        case CAMMSG_GETTHUMB:
            rc = GetThumb( hCam, (uint32_t)mp1, (char**)&pData);
            break;

        case CAMMSG_GETOBJECT:
            rc = GetObject( hCam, (uint32_t)mp1, (char**)&pData);
            break;

        case CAMMSG_SAVEOBJECT:
            rc = SaveObject( hCam, (CAMSaveObj*)mp1, mp2);
            break;

        case CAMMSG_DELETEOBJECT:
            if (fStop)
                rc = CAMERR_REQCANCELLED;
            else
                rc = DeleteObject( hCam, (uint32_t)mp1, 0);
            break;

        default:
            return (WinDefWindowProc( hwnd, msg, mp1, mp2));
    }

    // mp1 of the reply contains either the data or an errorcode;
    // mp2 contains the msg ID
    if (rc) {
        ClearStall( hCam);
        mpRtn = (MP)rc;
    }
    else
        mpRtn = (MP)pData;

    // if the reply can't be posted, free the data
    if (WinPostMsg( hReply, CAMMSG_REPLY, mpRtn, mp2) == FALSE) {
        camcli_error( "UsbWndProc - WinPostMsg() failed");
        if (pData)
            free( pData);
    }

    return (0);
}

/****************************************************************************/

// set the Usb thread's priority to ForegroundServer + 5;
// this should ensure it preempts the UI thread

void        ResetPriority( void)

{
    PTIB    ptib;
    PPIB    ppib;
    ULONG   priority;
    ULONG   class;
    LONG    delta;

    DosGetInfoBlocks( &ptib, &ppib);
    priority = ptib->tib_ptib2->tib2_ulpri;

    if (((priority & 0xff00) >> 8) == 4)
        class = 0;
    else
        class = 4;

    delta = 5 - (priority & 0xff);

    DosSetPriority( PRTYS_THREAD, class, delta, 0);
    return;
}

/****************************************************************************/

// This is intended as an optimization based on the original, published
// source code for usbcalls.  Whether it does anything in the current
// version is unknown since it's source hasn't been published.  Anyway...
//
// The buffer usbcalls passes to usbresmg has to be page-aligned.  If the
// app-supplied buffer isn't, usbcalls has to create its own buffer and
// then copy lots of data.  To avoid this, SaveObject() passes an unaligned
// buffer for the first 500 bytes of data, then uses a large buffer for the
// rest that is guaranteed to be page-aligned.

typedef struct _CAMReadData {
    uint32_t    junk[3];
    char        data[500];
} CAMReadData;

/***************************************************************************/

// This saves camera files to disk.  It's here rather than in camusb.c
// to avoid making the 'fStop' flag public & to keep that file PM-free.

uint32_t    SaveObject( CAMHandle hCam, CAMSaveObj * cso, MPARAM mp2)

{
    uint32_t        rtn;
    uint32_t        rc;
    uint32_t        size;
    uint32_t        alloc;
    uint32_t        cnt;
    uint32_t        wrote;
    uint32_t    	hFile;
    char *          data = 0;
    CAMReadData     crd;

do {
    hFile = cso->hFile;

    if (fStop) {
        rtn = CAMERR_REQCANCELLED;
        break;
    }

    rtn = ReadObjectBegin( hCam, cso->hObject, &size, &crd);
    if (rtn != PTP_RC_OK)
        break;

    cnt = (size <= sizeof(crd.data)) ? size : sizeof(crd.data);
    size -= cnt;

    rc = DosWrite( hFile, crd.data, cnt, (PULONG)&wrote);
    if (rc) {
        camcli_debug( "SaveObject - DosWrite  rc= 0x%04x", rc);
        rtn = CAMERR_WRITEFAILED;
        break;
    }

    if (size) {
        alloc = (size < 0x100000) ? size : 0x100000;

        // if this fails, we _should_ cancel the transaction
        rc = DosAllocMem( (void*)&data, alloc, PAG_COMMIT | PAG_READ | PAG_WRITE);
        if (rc) {
            camcli_debug( "SaveObject - DosAllocMem  rc= 0x%04x", rc);
            rtn = CAMERR_MALLOC;
            break;
        }
    }

    while (size) {
        if (size <= alloc) {
            cnt = size;
            cnt++;
        }
        else {
            cnt = alloc;
            WinPostMsg( hReply, CAMMSG_KEEPALIVE, 0, mp2);
        }

        if (fStop) {
            rtn = CancelRequest( hCam);
            if (!rtn)
                rtn = CAMERR_REQCANCELLED;
            break;
        }

        rtn = ReadObjectData( hCam, &cnt, data);
        if (rtn != PTP_RC_OK)
            break;

        rc = DosWrite( hFile, data, cnt, (PULONG)&wrote);
        if (rc) {
            camcli_debug( "SaveObject - DosWrite  rc= 0x%04x", rc);
            rtn = CAMERR_WRITEFAILED;
            break;
        }

        size -= cnt;
    }

    if (rtn != PTP_RC_OK)
        break;

    rtn = ReadObjectEnd( hCam);
    if (rtn == PTP_RC_OK)
        rtn = 0;

} while (0);

    if (data)
        DosFreeMem( data);

    return rtn;
}

/***************************************************************************/

