/****************************************************************************/
// camshow.c
//
//  - functions to fetch, convert, display, & rotate thumbnails
//  - sidebar window initialization & maintenance
//
/****************************************************************************/

/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Cameraderie.
 *
 * The Initial Developer of the Original Code is Richard L. Walsh.
 * 
 * Portions created by the Initial Developer are Copyright (C) 2006
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *  Heiko Nitzsche (gbm access code copied from gbmtest.c)
 *
 * ***** END LICENSE BLOCK ***** */

/****************************************************************************/

#include "camapp.h"
#include "camusb.h"
#include <stddef.h>
#include <fcntl.h>
#include <io.h>
#include "gbm.h"

/****************************************************************************/

#define CAM_MAGIC       0x6b7c8d9e
#define CAM_MAGICSTR    "0x6b7c8d9e"
#define CAM_MAGICJPG    "0x6b7c8d9e.JPG"

typedef struct _CAMFD {
    char *          pBuf;
    LONG            lSize;
    LONG            lPos;
} CAMFD;

// bitmap header incl. color table
// for 8 bit and lower images
#pragma pack(2)
typedef struct
{
   BITMAPINFOHEADER2 bmp2;
   RGB2 argb2Color[0x100];
} BMP8HEADER;

// for above 8 bit images
typedef struct
{
   BITMAPINFOHEADER2 bmp2;
} BMP24HEADER;
#pragma pack()

// used to create an NLS-appropriate date format
typedef struct _YMD {
    USHORT  y;
    USHORT  m;
    USHORT  d;
} YMD;


/****************************************************************************/
/*
void        InitShow( HWND hwnd);
void        TermShow( HWND hwnd);
HBITMAP     GetBlankBmp( void);
void        SetThumbWnd( HWND hwnd, CAMRecPtr pcr);
void        SetThumbWndTitle( HWND hwnd, char * pszTitle);
void        UpdateThumbWnd( HWND hwnd);
BOOL        RotateThumbBmp( CAMRecPtr pcr, ULONG ulRot);
void        FetchAllThumbnails( HWND hwnd);
*/

void        SetThumbWndInfo( HWND hSidebar, CAMRecPtr pcr);
void        InitThumbWndInfo( void);
HBITMAP     FetchOneThumbnail( CAMRecPtr pcr);
MRESULT _System SidebarSubProc( HWND hwnd, ULONG msg, MPARAM mp1, MPARAM mp2);

static HBITMAP  cam_create_bmp( const char *filename, ULONG ulRotate);
static HBITMAP  cam_rotate_bmp( HPS hps, HBITMAP hBmpIn,
                                PBITMAPINFOHEADER2 bmpInfo, ULONG ulRotate);

static int  GBMENTRY cam_open(const char *fn, int mode);
static int  GBMENTRY cam_create(const char *fn, int mode);
static void GBMENTRY cam_close(int fd);
static long GBMENTRY cam_lseek( int fd, long pos, int whence);
static int  GBMENTRY cam_read(int fd, void *buf, int len);
static int  GBMENTRY cam_write(int fd, const void *buf, int len);
static int  get_checked_internal_open_mode(int mode);
static int  get_checked_internal_lseek_mode(const int whence);
static int  isReadSupportedFormat(const int ft, const int bpp);

/****************************************************************************/

static PFNWP    pfnSidebar = 0;
static CAMFD    camFD = {0,0,0};
static HBITMAP  hbmpBlank = 0;
static char     szFormat[256];

/* mRot0 isn't used - it's just here to show what the default looks like
static MATRIXLF mRot0     = { 0x00010000, 0,          0,     //   1   0  0
                              0,          0x00010000, 0,     //   0   1  0
                              0,          0,          1 };   //   0   0  1
*/
static MATRIXLF mRot90    = { 0,          0x00010000, 0,     //   0   1  0
                              0xffff0000, 0,          0,     //  -1   0  0
                              120,        0,          1 };   // 120   0  1

static MATRIXLF mRot180   = { 0xffff0000, 0,          0,     //  -1   0  0
                              0,          0xffff0000, 0,     //   0  -1  0
                              160,        120,        1 };   // 160 120  1

static MATRIXLF mRot270   = { 0,          0xffff0000, 0,     //   0  -1  0
                              0x00010000, 0,          0,     //   1   0  0
                              0,          160,        1 };   //   0 160  1

// this identifies the ordering of year, month, & day when
// SetThumbWndInfo() constructs an NLS-appropriate date

static YMD      ymd[4] = {
                    {2,0,1},    // 0 MDY
                    {2,1,0},    // 1 DMY
                    {0,1,2},    // 2 YMD
                    {0,2,1}     // 3 YDM
                 };

static YMD *    pYMD;


static CAMNLS   nlsSidebar[] = {
    { IDC_EDITGRP,      BTN_EditGroups },
    { IDC_GRPFIRST,     GRP_Default },
    { 0,                0 }};


/****************************************************************************/

void        InitShow( HWND hwnd)

{
    HPS         hps;
    HWND        hSidebar;
    HWND        hTemp;
    SWP         swpInfo;

    hSidebar = WinWindowFromID( hwnd, IDC_SIDEBAR);
    LoadDlgStrings( hSidebar, nlsSidebar);
    InitThumbWndInfo();

    // the thumbnail window seems to ignore SM_SETHANDLE and won't
    // display new bitmaps if it is created with the bitmap style,
    // so we have to change the style & load the bitmap manually
    hps = WinGetPS( hSidebar);
    if (hps)
    {
        hbmpBlank = GpiLoadBitmap( hps, 0, IDB_THUMBBMP, 0, 0);
        WinReleasePS( hps);
    }

    hTemp = WinWindowFromID( hSidebar, IDC_BMP);
    WinSetWindowBits( hTemp, QWL_STYLE, SS_BITMAP, SS_TEXT | SS_BITMAP);
    WinSendMsg( hTemp, SM_SETHANDLE, (MP)hbmpBlank, 0);

    WinQueryWindowPos( hTemp, &swpInfo);

    hTemp = WinWindowFromID( hSidebar, IDC_ROTCC);
    swpInfo.y += 6;
    WinSetWindowPos( hTemp, 0, 8, swpInfo.y,
                     16, 16, SWP_MOVE | SWP_SIZE);

    hTemp = WinWindowFromID( hSidebar, IDC_ROTCW);
    swpInfo.x += swpInfo.cx - 16 - 8;
    WinSetWindowPos( hTemp, 0, swpInfo.x, swpInfo.y,
                     16, 16, SWP_MOVE | SWP_SIZE);

    InitGroups( hwnd);

    pfnSidebar = WinSubclassWindow( hSidebar, SidebarSubProc);

    gbm_init();
    gbm_io_setup( cam_open,
                  cam_create,
                  cam_close,
                  cam_lseek,
                  cam_read,
                  cam_write);

    return;
}

/****************************************************************************/

MRESULT _System SidebarSubProc( HWND hwnd, ULONG msg, MPARAM mp1, MPARAM mp2)
 
{

    if (msg == WM_WINDOWPOSCHANGED &&
        (((PSWP)mp1)->fl & SWP_SIZE) &&
        ((PSWP)mp1)[0].cy != ((PSWP)mp1)[1].cy) {

        HWND    hTemp;
        LONG    lChg;
        LONG    lExtra;
        SWP     aswp[5];

        lChg = ((PSWP)mp1)[0].cy - ((PSWP)mp1)[1].cy;

        hTemp = WinWindowFromID( hwnd, IDC_BMP);
        WinQueryWindowPos( hTemp, &aswp[0]);
        aswp[0].y += lChg;
        aswp[0].fl = SWP_MOVE;

        hTemp = WinWindowFromID( hwnd, IDC_TITLE);
        WinQueryWindowPos( hTemp, &aswp[1]);
        aswp[1].y += lChg;
        aswp[1].fl = SWP_MOVE;

        hTemp = WinWindowFromID( hwnd, IDC_ROTCW);
        WinQueryWindowPos( hTemp, &aswp[2]);
        aswp[2].y += lChg;
        aswp[2].fl = SWP_MOVE;

        hTemp = WinWindowFromID( hwnd, IDC_ROTCC);
        WinQueryWindowPos( hTemp, &aswp[3]);
        aswp[3].y += lChg;
        aswp[3].fl = SWP_MOVE;

        lExtra = RepositionGroupButtons( hwnd, ((PSWP)mp1)[0].cy);

        hTemp = WinWindowFromID( hwnd, IDC_INFO);
        WinQueryWindowPos( hTemp, &aswp[4]);
        aswp[4].y  = aswp[1].y;
        aswp[4].fl = SWP_MOVE | SWP_SIZE;
        if (lExtra > 0) {
            aswp[4].y -= lExtra;
            aswp[4].cy = lExtra;
        }
        else
            aswp[4].cy = 0;

        WinSetMultWindowPos( 0, aswp, 5);
    }

    return (pfnSidebar( hwnd, msg, mp1, mp2));
}

/****************************************************************************/

void        TermShow( HWND hwnd)

{
    HWND        hSidebar;


    if (camFD.pBuf) {
        free( camFD.pBuf);
        camFD.pBuf = 0;
    }

    gbm_deinit();

    hSidebar = WinWindowFromID( hwnd, IDC_SIDEBAR);
    WinSendDlgItemMsg( hSidebar, IDC_BMP, SM_SETHANDLE, 0, 0);

    if (hbmpBlank) {
        GpiDeleteBitmap( hbmpBlank);
        hbmpBlank = 0;
    }

    return;
}

/****************************************************************************/
/****************************************************************************/

HBITMAP     GetBlankBmp( void)

{
    return (hbmpBlank);
}

/****************************************************************************/

void        SetThumbWnd( HWND hwnd, CAMRecPtr pcr)

{
    HBITMAP     hbmp;
    HWND        hSidebar;
    HWND        hStatic;

    hSidebar = WinWindowFromID( hwnd, IDC_SIDEBAR);
    hStatic = WinWindowFromID( hSidebar, IDC_BMP);

    if (pcr->bmp)
        hbmp = pcr->bmp;
    else
        hbmp = hbmpBlank;

    WinSetWindowULong( hStatic, QWL_USER, (ULONG)pcr);
    SetThumbWndTitle( hwnd, pcr->title);
    WinSendMsg( hStatic, SM_SETHANDLE, (MP)hbmp, 0);
    SetThumbWndInfo( hSidebar, pcr);

    return;
}

/****************************************************************************/

void        SetThumbWndTitle( HWND hwnd, char * pszTitle)

{
    char *  ptr;
    char    szText[CCHMAXPATH];

    ptr = strrchr( pszTitle, '\\');
    if (ptr)
        ptr++;
    else {
        ptr = pszTitle;
        if (ptr[1] == ':')
            ptr += 2;
    }

    strcpy( szText, ptr);

    ptr = szText;
    while ((ptr = strpbrk( ptr, "\r\n")) != 0)
        strcpy( ptr, ptr + strspn( ptr, "\r\n"));

    WinSetDlgItemText( WinWindowFromID( hwnd, IDC_SIDEBAR), IDC_TITLE, szText);

    return;
}

/****************************************************************************/

void        SetThumbWndInfo( HWND hSidebar, CAMRecPtr pcr)

{
    USHORT  ausDate[3];
    char    szText[CCHMAXPATH];

    // put year, month, & day in the correct NLS order
    ausDate[pYMD->y] = pcr->date.year;
    ausDate[pYMD->m] = pcr->date.month;
    ausDate[pYMD->d] = pcr->date.day;

    sprintf( szText, szFormat, (int)pcr->size,
             ausDate[0], ausDate[1], ausDate[2], pcr->day,
             pcr->time.hours, pcr->time.minutes, pcr->time.seconds);

    WinSetDlgItemText( hSidebar, IDC_INFO, szText);

    return;
}

/****************************************************************************/

// constructs an NLS-apprpriate format string for SetThumbWndInfo();
// since GCC appears not to support %n$ to specify the order in which
// arguments should be used, we have to use our YMD structure to
// specify the ordering of year, month, & day

void        InitThumbWndInfo( void)

{
    ULONG   ul;
    char *  ptr;
    char    szSep[4];
    char    szWork[64];

    ptr = szFormat;

    // size
    LoadStringNoTilde( MNU_Size, szWork);
    ptr += sprintf( ptr, "\n%s:\t%%d", szWork);

    // date
    ul = (ULONG)PrfQueryProfileInt( HINI_USERPROFILE, "PM_National",
                                    "iDate", 2);
    if (ul > 3)
        ul = 2;
    pYMD = &ymd[ul];

    ul = PrfQueryProfileString( HINI_USERPROFILE, "PM_National",
                                "sDate", "-", szSep, sizeof(szSep));
    if (ul < 2 || ul > 3)
        strcpy( szSep, "-");

    LoadStringNoTilde( MNU_Date, szWork);
    ptr += sprintf( ptr, "\n%s:\t%%02d%s%%02d%s%%02d (%%s)",
                    szWork, szSep, szSep);

    // time
    ul = PrfQueryProfileString( HINI_USERPROFILE, "PM_National",
                                "sTime", ":", szSep, sizeof(szSep));
    if (ul < 2 || ul > 3)
        strcpy( szSep, ":");

    LoadStringNoTilde( MNU_Time, szWork);
    sprintf( ptr, "\n%s:\t%%02d%s%%02d%s%%02d", szWork, szSep, szSep);

    return;
}

/****************************************************************************/

// after deleting files, the current thumbnail may be invalid because the
// selection has changed or there aren't any records left;  this displays
// the t/n for the current record or the blank bitmap as appropriate

void        UpdateThumbWnd( HWND hwnd)

{
    HWND        	hSidebar;
    HWND        	hStatic;
    CAMRecPtr       pcr;
    CAMRecPtr       pcrOld;

do {
    hSidebar = WinWindowFromID( hwnd, IDC_SIDEBAR);
    hStatic = WinWindowFromID( hSidebar, IDC_BMP);

    pcrOld = (CAMRecPtr)WinQueryWindowULong( hStatic, QWL_USER);
    pcr = (CAMRecPtr)WinSendDlgItemMsg( hwnd, FID_CLIENT,
                    CM_QUERYRECORDEMPHASIS, (MP)CMA_FIRST, (MP)CRA_CURSORED);

    if (pcr && pcr != (CAMRecPtr)-1) {
        if (pcr != pcrOld) {
            WinSendDlgItemMsg( hwnd, FID_CLIENT, CM_SETRECORDEMPHASIS,
                               (MP)pcr, MPFROM2SHORT( TRUE, CRA_SELECTED));
            WinPostMsg( hwnd, CAMMSG_SHOWTHUMB, (MP)pcr, 0);
        }
        break;
    }

    WinSetWindowULong( hStatic, QWL_USER, 0);
    WinSetDlgItemText( hSidebar, IDC_TITLE, "");
    WinSetDlgItemText( hSidebar, IDC_INFO, "");
    UpdateStatus( hwnd, STS_NoFiles);
    WinSendMsg( hStatic, SM_SETHANDLE, (MP)hbmpBlank, 0);

} while (0);

    return;
}

/****************************************************************************/
/****************************************************************************/

BOOL        RotateThumbBmp( CAMRecPtr pcr, ULONG ulRot)

{
    BOOL        fRtn = FALSE;
    HBITMAP     hBmp;
    HBITMAP     hBmpRot;
    HDC         hdc = 0;
    HPS         hps = 0;
    SIZEL       sizl;
    BITMAPINFOHEADER2   bmih;

do {

    hBmp = pcr->bmp;
    if (hBmp == 0)
        break;

    memset( &bmih, 0, sizeof(bmih));
    bmih.cbFix = sizeof(bmih);

    if (!GpiQueryBitmapInfoHeader( hBmp, &bmih)) {
        printf( "RotateThumbBmp - GpiQueryBitmapInfoHeader failed\n");
        break;
    }

    hdc = DevOpenDC( 0, OD_MEMORY, "*", 0L, 0, 0);
    if (hdc == 0)
    {
        printf("DevOpenDC failed\n");
        break;
    }

    sizl.cx = ((bmih.cx >= bmih.cy) ? bmih.cx : bmih.cy);
    sizl.cy = sizl.cx;

    hps = GpiCreatePS( 0, hdc, &sizl, PU_PELS | GPIF_DEFAULT | GPIT_MICRO | GPIA_ASSOC);
    if (hps == 0)
    {
        printf("GpiCreatePS failed\n");
        break;
    }

    hBmpRot = cam_rotate_bmp( hps, hBmp, &bmih, ulRot);
    if (!hBmpRot)
        break;

    pcr->bmp = hBmpRot;
    GpiDeleteBitmap( hBmp);

    if (ulRot == STAT_ROT90)
        pcr->status = STAT_SETROTCC(pcr->status);
    else
        pcr->status = STAT_SETROTCW(pcr->status);

    fRtn = TRUE;

} while (0);

    if (hps) {
        GpiSetBitmap( hps, 0);
        GpiDestroyPS( hps);
    }
    if (hdc)
        DevCloseDC( hdc);

    return (fRtn);
}

/****************************************************************************/
/****************************************************************************/

void        FetchAllThumbnails( HWND hwnd)

{
    HWND        hCnr;
    CAMRecPtr   pcr = 0;

    hCnr = WinWindowFromID( hwnd, FID_CLIENT);

    pcr = (CAMRecPtr)WinSendMsg( hCnr, CM_QUERYRECORD, 0,
                                 MPFROM2SHORT( CMA_FIRST, CMA_ITEMORDER));
    if (!pcr || pcr == (CAMRecPtr)-1)
        return;

    do {
        if (!pcr->bmp && pcr->tnsize) {
            pcr->bmp = FetchOneThumbnail( pcr);
            if (pcr->bmp) {
                WinSendMsg( hCnr, CM_INVALIDATERECORD, (MP)&pcr,
                            MPFROM2SHORT( 1, CMA_NOREPOSITION));
                if (pcr->mrc.flRecordAttr & CRA_CURSORED)
                    WinPostMsg( hwnd, CAMMSG_SHOWTHUMB, (MP)pcr, 0);
            }
        }

        pcr = (CAMRecPtr)WinSendMsg( hCnr, CM_QUERYRECORD, (MP)pcr,
                                     MPFROM2SHORT( CMA_NEXT, CMA_ITEMORDER));

    } while (pcr && pcr != (CAMRecPtr)-1);

    return;
}

/****************************************************************************/

HBITMAP     FetchOneThumbnail( CAMRecPtr pcr)

{
    HBITMAP     hbmpRtn = 0;
    char *      ptr;
    char        szText[32];

do {

    // get the thumbnail from the camera 
    if (DispatchUsbMsg( CAMMSG_GETTHUMB, pcr->hndl, (void**)&camFD.pBuf) ||
        !camFD.pBuf) {
        camcli_error( "Could not get thumbnail for handle %d", (int)pcr->hndl);
        break;
    }

    // turn the data into a bitmap

    // setup this unfortunate method of passing data to gbm
    camFD.lSize = (LONG)pcr->tnsize;
    camFD.lPos  = 0;

    // our gbm_io_open() replacement recognizes a camera
    // request by this (hopefully unique) filename
    if (pcr->tnfmtnbr == PTP_OFC_JFIF ||
        pcr->tnfmtnbr == PTP_OFC_EXIF_JPEG)
        strcpy( szText, CAM_MAGICJPG);
    else {
        strcpy( szText, CAM_MAGICSTR);
        ptr = strrchr( pcr->title, '.');
        if (ptr)
            strcat( szText, ptr);
    }

    // create the bitmap and rotate it if needed
    hbmpRtn = cam_create_bmp( szText, STAT_GETROT(pcr->status));

    // our gbm_io_close() replacement should already have freed
    // the buffer returned by GetThumb(), but check to make sure
    if (camFD.pBuf) {
        free( camFD.pBuf);
        camFD.pBuf = 0;
    }

} while (0);

    return (hbmpRtn);
}

/****************************************************************************/

static HBITMAP cam_create_bmp( const char *filename, ULONG ulRotate)
{
    int         i;
    int         ftr;
    int         dataSize;
    GBM         gbm;
    GBMRGB      gbmrgb[0x100];
    BMP8HEADER  header8BMP;
    BMP24HEADER header24BMP;
    SIZEL       sizl;
    BITMAPINFO2 *bmpInfo;

    int         fd = -1;
    byte  *     data = NULL;
    HBITMAP     hBmp = 0;
    HDC         hdc = 0;
    HPS         hps = 0;

do {
    /* check if file format is supported */
    if (gbm_guess_filetype(filename, &ftr) != GBM_ERR_OK)
    {
        printf("Input format is not supported.\n");
        break;
    }
 
    /* Read bitmap file */
 
    /* open bitmap file */
    fd = gbm_io_open(filename, GBM_O_RDONLY);
    if (fd < 0)
    {
        printf("FAILED, Cannot open file.\n");
        break;
    }
 
    /* read bitmap header */
    if (gbm_read_header(filename, fd, ftr, &gbm, "") != GBM_ERR_OK)
    {
        printf("FAILED, Cannot read file header.\n");
        break;
    }
 
    /* check whether we can support the provided color depth (restrict to algorithms) */
    if (!isReadSupportedFormat(ftr, gbm.bpp))
    {
        printf("FAILED: Input format color depth is not supported.\n");
        break;
    }
 
    /* read bitmap palette */
    memset(gbmrgb, 0, sizeof(gbmrgb));
    if (gbm_read_palette(fd, ftr, &gbm, gbmrgb) != GBM_ERR_OK)
    {
        printf("FAILED, Cannot read bitmap palette.\n");
        break;
    }
 
    /* calc needed memory and load bitmap data */
    dataSize = (((gbm.w * gbm.bpp + 31) / 32) * 4) * gbm.h;
    if ((data = (byte *) malloc(dataSize)) == NULL)
    {
        printf("FAILED, Out of memory.\n");
        break;
    }
    memset(data, 0, dataSize);
 
    /* read bitmap data */
    if (gbm_read_data(fd, ftr, &gbm, data) != GBM_ERR_OK)
    {
        printf("FAILED, Cannot read bitmap data.\n");
        break;
    }
 
    gbm_io_close(fd);
    fd = -1;
 
    /* create bitmap header */
    if (gbm.bpp <= 8)
    {
        bmpInfo = (PBITMAPINFO2)&header8BMP;
        memset( &header8BMP, 0, sizeof(BMP8HEADER));

        header8BMP.bmp2.cbFix           = sizeof(BITMAPINFOHEADER2);
        header8BMP.bmp2.cx              = gbm.w;
        header8BMP.bmp2.cy              = gbm.h;
        header8BMP.bmp2.cPlanes         = 1;                /*  Number of bit planes. */
        header8BMP.bmp2.cBitCount       = gbm.bpp;          /*  Number of bits per pel within a plane. */
        header8BMP.bmp2.ulCompression   = BCA_UNCOMP;       /*  Compression scheme used to store the bit map. */
        header8BMP.bmp2.cbImage         = dataSize;         /*  Length of bit-map storage data, in bytes. */

        /* copy color table */
        if ((gbm.bpp == 1) &&
            ((gbmrgb[1].b == 0)    && (gbmrgb[1].r == 0)    && (gbmrgb[1].g == 0)  &&
             (gbmrgb[0].b == 0xff) && (gbmrgb[0].r == 0xff) && (gbmrgb[0].g == 0xff)))
        {
            /* special handling of BW bitmaps due to the way PM works */
            header8BMP.argb2Color[1].bBlue     = 0xff;
            header8BMP.argb2Color[1].bRed      = 0xff;
            header8BMP.argb2Color[1].bGreen    = 0xff;
        }
        else
            for (i = 0; i < 256; i++)
            {
                header8BMP.argb2Color[i].bBlue     = gbmrgb[i].b;
                header8BMP.argb2Color[i].bRed      = gbmrgb[i].r;
                header8BMP.argb2Color[i].bGreen    = gbmrgb[i].g;
            }
    }
    else
    {
        bmpInfo = (PBITMAPINFO2) &header24BMP;
        memset( &header24BMP, 0, sizeof(BMP24HEADER));

        header24BMP.bmp2.cbFix           = sizeof(BITMAPINFOHEADER2);
        header24BMP.bmp2.cx              = gbm.w;
        header24BMP.bmp2.cy              = gbm.h;
        header24BMP.bmp2.cPlanes         = 1;                /*  Number of bit planes. */
        header24BMP.bmp2.cBitCount       = gbm.bpp;          /*  Number of bits per pel within a plane. */
        header24BMP.bmp2.ulCompression   = BCA_UNCOMP;       /*  Compression scheme used to store the bit map. */
        header24BMP.bmp2.cbImage         = dataSize;         /*  Length of bit-map storage data, in bytes. */
    }

   /* create system bitmap */

    /* create a memory context */
    hdc = DevOpenDC( 0, OD_MEMORY, "*", 0L, 0, 0);
    if (hdc == 0)
    {
        printf("FAILED, DevOpenDC\n");
        break;
    }

    sizl.cx = ((bmpInfo->cx >= bmpInfo->cy) ? bmpInfo->cx : bmpInfo->cy);
    sizl.cy = sizl.cx;

    /* create presentation space */
    hps = GpiCreatePS( 0, hdc, &sizl, PU_PELS | GPIF_DEFAULT | GPIT_MICRO | GPIA_ASSOC);
    if (hps == 0)
    {
        printf("FAILED, GpiCreatePS\n");
        break;
    }

    /* create the bitmap */
    hBmp = GpiCreateBitmap( hps, (PBITMAPINFOHEADER2)bmpInfo,
                            CBM_INIT, (PBYTE) data, bmpInfo );

    if (ulRotate) {
        HBITMAP hBmpRot = cam_rotate_bmp( hps, hBmp,
                                          (PBITMAPINFOHEADER2)bmpInfo,
                                          ulRotate);
        if (hBmpRot) {
            GpiDeleteBitmap( hBmp);
            hBmp = hBmpRot;
        }
    }

} while (0);

    if (hps) {
        GpiSetBitmap( hps, 0);
        GpiDestroyPS( hps);
    }
    if (hdc)
        DevCloseDC( hdc);
    if (data)
        free( data);
    if (fd >= 0)
        gbm_io_close( fd);

    return hBmp;
}

/****************************************************************************/

static HBITMAP cam_rotate_bmp( HPS hps, HBITMAP hBmpIn,
                               PBITMAPINFOHEADER2 bmpInfo, ULONG ulRotate)
{
    MATRIXLF *  pRot;
    HBITMAP     hBmpOut = 0;
    HBITMAP     hBmpWork = 0;
    POINTL      aptl[4];

do {
    memset( aptl, 0, sizeof(aptl));
    aptl[1].x = bmpInfo->cx;
    aptl[1].y = bmpInfo->cy;
    aptl[3].x = bmpInfo->cx;
    aptl[3].y = bmpInfo->cy;

    if (ulRotate == STAT_ROT0)
        break;

    if (ulRotate == STAT_ROT180) {
        pRot = &mRot180;
        mRot180.lM31 = bmpInfo->cx;
        mRot180.lM32 = bmpInfo->cy;
    }
    else {
        bmpInfo->cx = aptl[1].y;
        bmpInfo->cy = aptl[1].x;

        if (ulRotate == STAT_ROT270) {
            pRot = &mRot270;
            mRot270.lM32 = bmpInfo->cy;
        }
        else {
            pRot = &mRot90;
            mRot90.lM31 = bmpInfo->cx;
        }
    }

    GpiSetBitmap( hps, 0);

    hBmpWork = GpiCreateBitmap( hps, bmpInfo, 0, 0, 0);
    if (hBmpWork == 0) {
        printf( "GpiCreateBitmap failed\n");
        break;
    }

    if (GpiSetBitmap( hps, hBmpWork) == HBM_ERROR) {
        printf( "GpiSetBitmap failed\n");
        break;
    }

    if (!GpiSetModelTransformMatrix( hps, 9, pRot, TRANSFORM_REPLACE)) {
        printf( "Rotate failed\n");
        break;
    }

    if (GpiWCBitBlt( hps, hBmpIn, 4, aptl, ROP_SRCCOPY,
                     BBO_IGNORE) == GPI_ERROR) {
        printf( "GpiWCBitBlt failed\n");
        break;
    }

    hBmpOut = hBmpWork;

} while (0);

    if (hBmpOut == 0) {
        if (hBmpWork)
            GpiDeleteBitmap( hBmpWork);
    }

    return hBmpOut;
}

/****************************************************************************/
/****************************************************************************/

static int GBMENTRY cam_open(const char *fn, int mode)

{
    if (strtoul( fn, 0, 0) != CAM_MAGIC)
       return open( fn, get_checked_internal_open_mode( mode));

    if (camFD.pBuf == 0)
        return -1;

    camFD.lPos = 0;

    return (int)CAM_MAGIC;
}

static int GBMENTRY cam_create(const char *fn, int mode)

{
    if (strtoul( fn, 0, 0) == CAM_MAGIC)
        return -1;

    return open(fn, O_CREAT | O_TRUNC | get_checked_internal_open_mode(mode), S_IREAD | S_IWRITE);
}

static void GBMENTRY cam_close(int fd)

{ 
    if (fd != (int)CAM_MAGIC)
       close( fd);
    else {
        free( camFD.pBuf);
        camFD.pBuf = 0;
    }

    return;
}

static long GBMENTRY cam_lseek( int fd, long pos, int whence)

{
    LONG    lCur;

    if (fd != (int)CAM_MAGIC)
        return lseek( fd, pos, get_checked_internal_lseek_mode(whence));

    switch (whence) {

        case GBM_SEEK_SET:
            lCur = pos;
            break;

        case GBM_SEEK_CUR:
            lCur = camFD.lPos + pos;
            break;

        case GBM_SEEK_END:
            lCur = camFD.lSize + pos;
            break;

        default:
            lCur = -1;
    }

    if (lCur < 0 || lCur > camFD.lSize)
        lCur = -1;
    else
        camFD.lPos = lCur;

    return lCur;
}

static int GBMENTRY cam_read(int fd, void *buf, int len)

{ 
    LONG    lRead;

    if (!buf || len <= 0)
        return -1;

    if (fd != (int)CAM_MAGIC)
        return read(fd, buf, len);

    if (camFD.lPos + len > camFD.lSize)
        lRead = camFD.lSize - camFD.lPos;
    else
        lRead = len;

    memcpy( buf, &camFD.pBuf[camFD.lPos], lRead);
    camFD.lPos += lRead;

    return lRead;
}

static int GBMENTRY cam_write(int fd, const void *buf, int len)

{
    if (fd != (int)CAM_MAGIC)
        return write(fd, buf, len);

    return -1;
}

/****************************************************************************/

/* map supported GBM_O_* defines to compiler specific ones */
static int get_checked_internal_open_mode(int mode)
{
   const static int supported_open_modes = GBM_O_RDONLY |
                                           GBM_O_WRONLY |
                                           GBM_O_RDWR   |
                                           GBM_O_EXCL   |
                                           GBM_O_NOINHERIT;

   /* internal, compiler specific open mode */
   int open_mode = 0;

   /* map supported public defines to compiler specific ones */

   /* check if only supported modes are provided */
   if (!(mode & supported_open_modes))
   {
      return 0xffffffff;
   }

   /* mask external binary mode bit */
   mode &= ~GBM_O_BINARY;

   if (mode & GBM_O_RDONLY)
   {
      open_mode |= O_RDONLY;
   }
   else if (mode & GBM_O_WRONLY)
   {
      open_mode |= O_WRONLY;
   }
   else if (mode & GBM_O_RDWR)
   {
      open_mode |= O_RDWR;
   }
   else if (mode & GBM_O_EXCL)
   {
      open_mode |= O_EXCL;
   }
   else
   {
      return 0xffffffff;
   }

   if (mode & GBM_O_NOINHERIT)
   {
      open_mode |= O_NOINHERIT;
   }

   /* force binary mode if necessary */
   #ifdef O_BINARY
     open_mode |= O_BINARY;
   #endif

   return open_mode;
}

/* map supported GBM_SEEK_* defines to compiler specific ones */
static int get_checked_internal_lseek_mode( const int whence)

{
   int internal_whence = -1;

   switch(whence)
   {
      case GBM_SEEK_SET:
         internal_whence = SEEK_SET;
         break;

      case GBM_SEEK_CUR:
         internal_whence = SEEK_CUR;
         break;

      case GBM_SEEK_END:
         internal_whence = SEEK_END;
         break;

      default:
         /* as we provide an illegal whence value, the OS takes care
          * for correct error reporting
          */
         break;
   }

   return internal_whence;
}

/****************************************************************************/

static int isReadSupportedFormat( const int ft, const int bpp)

{
   GBMFT readFormat;
   int   readFormatFlag = 0;

   switch (bpp)
   {
      case 64: readFormatFlag = GBM_FT_R64; break;
      case 48: readFormatFlag = GBM_FT_R48; break;
      case 32: readFormatFlag = GBM_FT_R32; break;
      case 24: readFormatFlag = GBM_FT_R24; break;
      case  8: readFormatFlag = GBM_FT_R8;  break;
      case  4: readFormatFlag = GBM_FT_R4;  break;
      case  1: readFormatFlag = GBM_FT_R1;  break;
      default: readFormatFlag = 0;          break;
   }

   if (gbm_query_filetype( ft, &readFormat) == GBM_ERR_OK &&
       (readFormat.flags & readFormatFlag))
       return 1;

   return 0;
}

/****************************************************************************/

